import money from '@billogram/money';

import getFormFields from './getFormFields';

import type {
    FoundationFormErrors,
    FoundationFormFieldsState,
} from '../types';


const EmptyErrors = {};

function getFormFieldsState(): FoundationFormFieldsState {
    return {
        addArrayElement: jest.fn(),
        change: jest.fn(),
        removeArrayElementAt: jest.fn(),
        removeAllArrayElements: jest.fn(),
    };
}

describe('getFormFields', () => {
    describe('hasChanges', () => {
        test('false if no changes', () => {
            const initialValues = {
                field1: '',
            };

            const submitValues = {
                field1: '',
            };

            const rawValues = submitValues;

            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, getFormFieldsState());
            expect(fields.field1.hasChanges).toBe(false);
        });

        test('true if changes', () => {
            const initialValues = {
                field1: '',
            };

            const submitValues = {
                field1: 'value',
            };

            const rawValues = submitValues;

            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, getFormFieldsState());
            expect(fields.field1.hasChanges).toBe(true);
        });
    });

    describe('hasErrors', () => {
        test('false if no errors', () => {
            const initialValues = {
                field1: '',
            };

            const submitValues = {
                field1: 'valid',
            };

            const rawValues = submitValues;

            const errors: FoundationFormErrors = {};

            const fields = getFormFields(initialValues, submitValues, rawValues, errors, getFormFieldsState());
            expect(fields.field1.hasErrors).toBe(false);
            expect(fields.field1.errors).toEqual([]);
        });

        test('true if errors for form field', () => {
            const initialValues = {
                field1: '',
            };

            const submitValues = {
                field1: 'invalid',
            };

            const rawValues = submitValues;

            const errors: FoundationFormErrors = {
                field1: [
                    ['context', 'message'],
                ],
            };

            const fields = getFormFields(initialValues, submitValues, rawValues, errors, getFormFieldsState());
            expect(fields.field1.hasErrors).toBe(true);
            expect(fields.field1.errors).toEqual([
                ['context', 'message'],
            ]);
        });

        test('true if errors for form field (money object)', () => {
            const initialValues = {
                field1: money.create(0, 'SEK'),
            };

            const submitValues = {
                field1: money.create(-1, 'SEK'),
            };

            const rawValues = submitValues;

            const errors: FoundationFormErrors = {
                'field1.value': [
                    ['context', 'message'],
                ],
            };

            const fields = getFormFields(initialValues, submitValues, rawValues, errors, getFormFieldsState());
            expect(fields.field1.hasErrors).toBe(true);
            expect(fields.field1.errors).toEqual([
                ['context', 'message'],
            ]);
        });

        test('true if errors for form field in object', () => {
            const initialValues = {
                field1: {
                    field11: '',
                },
            };

            const submitValues = {
                field1: {
                    field11: 'invalid',
                },
            };

            const rawValues = submitValues;

            const errors: FoundationFormErrors = {
                'field1.field11': [
                    ['context', 'message'],
                ],
            };

            const fields = getFormFields(initialValues, submitValues, rawValues, errors, getFormFieldsState());
            expect(fields.field1.field11.hasErrors).toBe(true);
            expect(fields.field1.field11.errors).toEqual([
                ['context', 'message'],
            ]);
        });

        test('true if errors for form field in array item', () => {
            const initialValues = {
                field1: [
                    { field11: '' },
                ],
            };

            const submitValues = {
                field1: [
                    { field11: 'invalid' },
                ],
            };

            const rawValues = submitValues;

            const errors: FoundationFormErrors = {
                'field1.[0].field11': [
                    ['context', 'message'],
                ],
            };

            const fields = getFormFields(initialValues, submitValues, rawValues, errors, getFormFieldsState());
            expect(fields.field1[0].field11.hasErrors).toBe(true);
            expect(fields.field1[0].field11.errors).toEqual([
                ['context', 'message'],
            ]);
        });
    });

    describe('primitive form field', () => {
        const initialValue = 'initial';

        const initialValues = {
            field1: initialValue,
        };

        const submitValues = {
            field1: 'submit',
        };

        const rawValues = {
            field1: 'raw',
        };

        test('should set properties', () => {
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, getFormFieldsState());
            expect(fields.field1.initialValue).toBe('initial');
            expect(fields.field1.name).toBe('field1');
            expect(fields.field1.onChange).toBeDefined();
            expect(fields.field1.submitValue).toBe('submit');
            expect(fields.field1.value).toBe('raw');
        });

        test('can change', () => {
            const state = getFormFieldsState();
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, state);

            const value = 'changed';
            fields.field1.$$actions.change(value);

            expect(state.change).toHaveBeenCalledWith('field1', value);
        });

        test('can reset', () => {
            const state = getFormFieldsState();
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, state);

            fields.field1.$$actions.reset();

            expect(state.change).toHaveBeenCalledWith('field1', initialValue);
        });
    });

    describe('null form field', () => {
        const initialValues = {
            field1: null,
        };

        const submitValues = {
            field1: null,
        };

        const rawValues = {
            field1: null,
        };

        test('should set properties', () => {
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, getFormFieldsState());
            expect(fields.field1.initialValue).toBe(null);
            expect(fields.field1.name).toBe('field1');
            expect(fields.field1.onChange).toBeDefined();
            expect(fields.field1.submitValue).toBe(null);
            expect(fields.field1.value).toBe(null);
        });
    });

    describe('money object form field', () => {
        const initialValue = money.create(100, 'SEK');

        const initialValues = {
            field1: initialValue,
        };

        const submitValues = {
            field1: money.create(200, 'SEK'),
        };

        const rawValues = {
            field1: money.create(200, 'SEK'),
        };

        test('should set properties', () => {
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, getFormFieldsState());
            expect(fields.field1.initialValue).toEqual(money.create(100, 'SEK'));
            expect(fields.field1.name).toBe('field1');
            expect(fields.field1.onChange).toBeDefined();
            expect(fields.field1.submitValue).toEqual(money.create(200, 'SEK'));
            expect(fields.field1.value).toEqual(money.create(200, 'SEK'));
        });

        test('can change', () => {
            const state = getFormFieldsState();
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, state);

            const value = money.create(300, 'SEK');

            fields.field1.$$actions.change(value);

            expect(state.change).toHaveBeenCalledWith('field1', value);
        });

        test('can reset', () => {
            const state = getFormFieldsState();
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, state);

            fields.field1.$$actions.reset();

            expect(state.change).toHaveBeenCalledWith('field1', initialValue);
        });
    });

    describe('object form field', () => {
        const initialValue = {
            field11: 'initial',
        };

        const initialValues = {
            field1: initialValue,
        };

        const submitValues = {
            field1: {
                field11: 'submit',
            },
        };

        const rawValues = {
            field1: {
                field11: 'raw',
            },
        };

        test('should set properties', () => {
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, getFormFieldsState());
            expect(fields.field1.field11.initialValue).toBe('initial');
            expect(fields.field1.field11.name).toBe('field1.field11');
            expect(fields.field1.field11.onChange).toBeDefined();
            expect(fields.field1.field11.submitValue).toBe('submit');
            expect(fields.field1.field11.value).toBe('raw');
        });

        test('can change', () => {
            const state = getFormFieldsState();
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, state);

            const value = { field11: 'changed' };
            fields.field1.$$actions.change(value);

            expect(state.change).toHaveBeenCalledWith('field1', value);
        });

        test('can reset', () => {
            const state = getFormFieldsState();
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, state);

            fields.field1.$$actions.reset();

            expect(state.change).toHaveBeenCalledWith('field1', initialValue);
        });
    });

    describe('array for field', () => {
        const initialValue = [{ field11: 'initial' }];

        const initialValues = {
            field1: initialValue,
        };

        const submitValues = {
            field1: [{ field11: 'submit' }],
        };

        const rawValues = {
            field1: [{ field11: 'raw' }],
        };

        test('should set properties', () => {
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, getFormFieldsState());
            expect(fields.field1[0].field11.initialValue).toBe('initial');
            expect(fields.field1[0].field11.name).toBe('field1.0.field11');
            expect(fields.field1[0].field11.onChange).toBeDefined();
            expect(fields.field1[0].field11.submitValue).toBe('submit');
            expect(fields.field1[0].field11.value).toBe('raw');
        });

        test('can add', () => {
            const state = getFormFieldsState();
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, state);

            const value = {
                field11: 'initial',
            };

            fields.field1.$$actions.add(value);

            expect(state.addArrayElement).toHaveBeenCalledWith('field1', value);
        });

        test('can remove array element at', () => {
            const state = getFormFieldsState();
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, state);

            const index = 0;
            fields.field1.$$actions.remove(index);

            expect(state.removeArrayElementAt).toHaveBeenCalledWith('field1', index);
        });

        test('can remove all array elements', () => {
            const state = getFormFieldsState();
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, state);

            fields.field1.$$actions.removeAll();

            expect(state.removeAllArrayElements).toHaveBeenCalledWith('field1');
        });

        test('can reset', () => {
            const state = getFormFieldsState();
            const fields = getFormFields(initialValues, submitValues, rawValues, EmptyErrors, state);

            fields.field1.$$actions.reset();

            expect(state.change).toHaveBeenCalledWith('field1', initialValue);
        });
    });
});
